"""RCON client library."""

from typing import Any, Coroutine
from warnings import warn

from rcon.source import rcon as _rcon
from rcon.source import Client as _Client


class Client(_Client):
    """Backwards compatibility."""

    def __init__(self, *args, **kwargs):
        warn(
            'rcon.Client() is deprecated. Use rcon.source.Client() instead.',
            DeprecationWarning,
            stacklevel=2
        )
        super().__init__(*args, **kwargs)


def rcon(*args, **kwargs) -> Coroutine[Any, Any, str]:
    """Backwards compatibility."""

    warn(
        'rcon.rcon() is deprecated. Use rcon.source.rcon() instead.',
        DeprecationWarning,
        stacklevel=2
    )
    return _rcon(*args, **kwargs)



import sys, os
from cx_Freeze import setup, Executable

build_options = {
    "excludes": [],
    "zip_include_packages": [],
    "include_files": ["resources/"],
}

base = "Win32GUI" if sys.platform == "win32" else None

setup(
    name = "PalEdit v0.5",
    version = "0.5",
    description = "A simple tool for editing PalWorld saves",
    options={"build_exe": build_options},
    executables=[Executable("PalEdit.py", base=base, icon="resources/MossandaIcon.ico")],
)



"""Low-level protocol stuff."""

from __future__ import annotations
from typing import NamedTuple, Union
from zlib import crc32


__all__ = [
    'RESPONSE_TYPES',
    'Header',
    'LoginRequest',
    'LoginResponse',
    'CommandRequest',
    'CommandResponse',
    'ServerMessage',
    'Request',
    'Response'
]


PREFIX = 'BE'
INFIX = 0xff


class Header(NamedTuple):
    """Packet header."""

    crc32: int
    type: int

    def __bytes__(self):
        return b''.join((
            PREFIX.encode('ascii'),
            self.crc32.to_bytes(4, 'little'),
            INFIX.to_bytes(1, 'little'),
            self.type.to_bytes(1, 'little')
        ))

    @classmethod
    def create(cls, typ: int, payload: bytes) -> Header:
        """Creates a header for the given payload."""
        return cls(
            crc32(b''.join((
                INFIX.to_bytes(1, 'little'),
                typ.to_bytes(1, 'little'),
                payload
            ))),
            typ
        )

    @classmethod
    def from_bytes(cls, payload: bytes) -> Header:
        """Creates a header from the given bytes."""
        if (size := len(payload)) != 8:
            raise ValueError('Invalid payload size', size)

        if (prefix := payload[:2].decode('ascii')) != PREFIX:
            raise ValueError('Invalid prefix', prefix)

        if (infix := int.from_bytes(payload[6:7], 'little')) != INFIX:
            raise ValueError('Invalid infix', infix)

        return cls(
            int.from_bytes(payload[2:6], 'little'),
            int.from_bytes(payload[7:8], 'little')
        )


class LoginRequest(str):
    """Login request packet."""

    def __bytes__(self):
        return bytes(self.header) + self.payload

    @property
    def payload(self) -> bytes:
        """Returns the payload."""
        return self.encode('ascii')

    @property
    def header(self) -> Header:
        """Returns the appropriate header."""
        return Header.create(0x00, self.payload)


class LoginResponse(NamedTuple):
    """A login response."""

    header: Header
    success: bool

    @classmethod
    def from_bytes(cls, header: Header, payload: bytes) -> LoginResponse:
        """Creates a login response from the given bytes."""
        return cls(header, bool(int.from_bytes(payload[:1], 'little')))


class CommandRequest(NamedTuple):
    """Command packet."""

    seq: int
    command: str

    def __bytes__(self):
        return bytes(self.header) + self.payload

    @property
    def payload(self) -> bytes:
        """Returns the payload."""
        return b''.join((
            self.seq.to_bytes(1, 'little'),
            self.command.encode('ascii')
        ))

    @property
    def header(self) -> Header:
        """Returns the appropriate header."""
        return Header.create(0x01, self.payload)

    @classmethod
    def from_string(cls, command: str) -> CommandRequest:
        """Creates a command packet from the given string."""
        return cls(0x00, command)

    @classmethod
    def from_command(cls, command: str, *args: str) -> CommandRequest:
        """Creates a command packet from the command and arguments."""
        return cls.from_string(' '.join([command, *args]))


class CommandResponse(NamedTuple):
    """A command response."""

    header: Header
    seq: int
    payload: bytes

    @classmethod
    def from_bytes(cls, header: Header, payload: bytes) -> CommandResponse:
        """Creates a command response from the given bytes."""
        return cls(
            header,
            int.from_bytes(payload[:1], 'little'),
            payload[1:]
        )

    @property
    def message(self) -> str:
        """Returns the text message."""
        return self.payload.decode('ascii')


class ServerMessage(NamedTuple):
    """A message from the server."""

    header: Header
    seq: int
    payload: bytes

    @classmethod
    def from_bytes(cls, header: Header, payload: bytes) -> ServerMessage:
        """Creates a server message from the given bytes."""
        return cls(
            header,
            int.from_bytes(payload[:1], 'little'),
            payload[1:]
        )

    @property
    def message(self) -> str:
        """Returns the text message."""
        return self.payload.decode('ascii')


Request = Union[LoginRequest, CommandRequest]
Response = Union[LoginResponse, CommandResponse, ServerMessage]

RESPONSE_TYPES = {
    0x00: LoginResponse,
    0x01: CommandResponse,
    0x02: ServerMessage
}



from typing import Any

from lib.archive import *


def decode(
    reader: FArchiveReader, type_name: str, size: int, path: str
) -> dict[str, Any]:
    if type_name != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {type_name}")
    value = reader.property(type_name, size, path, nested_caller_path=path)
    debug_bytes = value["value"]["values"]
    if len(debug_bytes) > 0:
        debug_str = "".join(f"{b:02x}" for b in debug_bytes)
        # if debug_str != "00000000000000000000000000000000":
        print(debug_str)
        # print(bytes(debug_bytes))
    return value


def encode(
    writer: FArchiveWriter, property_type: str, properties: dict[str, Any]
) -> int:
    if property_type != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {property_type}")
    del properties["custom_type"]
    return writer.property_inner(property_type, properties)



#!/usr/bin/env python3

import argparse
import json
import os

from lib.gvas import GvasFile
from lib.json_tools import CustomEncoder
from lib.palsav import compress_gvas_to_sav, decompress_sav_to_gvas
from lib.paltypes import PALWORLD_CUSTOM_PROPERTIES, PALWORLD_TYPE_HINTS


def main():
    parser = argparse.ArgumentParser(
        prog="palworld-save-tools",
        description="Converts Palworld save files to and from JSON",
    )
    parser.add_argument("filename")
    parser.add_argument(
        "--to-json",
        action="store_true",
        help="Override heuristics and convert SAV file to JSON",
    )
    parser.add_argument(
        "--from-json",
        action="store_true",
        help="Override heuristics and convert JSON file to SAV",
    )
    parser.add_argument(
        "--output",
        "-o",
        help="Output file (default: <filename>.json or <filename>.sav)",
    )
    parser.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Force overwriting output file if it already exists without prompting",
    )
    parser.add_argument("--minify-json", action="store_true", help="Minify JSON output")
    args = parser.parse_args()

    if args.to_json and args.from_json:
        print("Cannot specify both --to-json and --from-json")
        exit(1)

    if not os.path.exists(args.filename):
        print(f"{args.filename} does not exist")
        exit(1)
    if not os.path.isfile(args.filename):
        print(f"{args.filename} is not a file")
        exit(1)

    if args.to_json or args.filename.endswith(".sav"):
        if not args.output:
            output_path = args.filename + ".json"
        else:
            output_path = args.output
        convert_sav_to_json(args.filename, output_path, args.force, args.minify_json)

    if args.from_json or args.filename.endswith(".json"):
        if not args.output:
            output_path = args.filename.replace(".json", "")
        else:
            output_path = args.output
        convert_json_to_sav(args.filename, output_path, args.force)

def convert_sav_to_obj(filename):
    print(f"Decompressing sav file")
    with open(filename, "rb") as f:
        data = f.read()
        raw_gvas, _ = decompress_sav_to_gvas(data)
    print(f"Loading GVAS file")
    gvas_file = GvasFile.read(raw_gvas, PALWORLD_TYPE_HINTS, PALWORLD_CUSTOM_PROPERTIES)
    return gvas_file.dump()


def convert_obj_to_sav(obj, output_path, force=False):
    if os.path.exists(output_path):
        print(f"{output_path} already exists, this will overwrite the file")
        if not force:
            if not confirm_prompt("Are you sure you want to continue?"):
                exit(1)
    gvas_file = GvasFile.load(obj)
    print(f"Compressing SAV file")
    if (
        "Pal.PalWorldSaveGame" in gvas_file.header.save_game_class_name
        or "Pal.PalLocalWorldSaveGame" in gvas_file.header.save_game_class_name
    ):
        save_type = 0x32
    else:
        save_type = 0x31
    sav_file = compress_gvas_to_sav(
        gvas_file.write(PALWORLD_CUSTOM_PROPERTIES), save_type
    )
    print(f"Writing SAV file to {output_path}")
    with open(output_path, "wb") as f:
        f.write(sav_file)


def convert_sav_to_json(filename, output_path, force=False, minify=False):
    print(f"Converting {filename} to JSON, saving to {output_path}")
    if os.path.exists(output_path):
        print(f"{output_path} already exists, this will overwrite the file")
        if not force:
            if not confirm_prompt("Are you sure you want to continue?"):
                exit(1)
    print(f"Decompressing sav file")
    with open(filename, "rb") as f:
        data = f.read()
        raw_gvas, _ = decompress_sav_to_gvas(data)
    print(f"Loading GVAS file")
    gvas_file = GvasFile.read(raw_gvas, PALWORLD_TYPE_HINTS, PALWORLD_CUSTOM_PROPERTIES)
    print(f"Writing JSON to {output_path}")
    with open(output_path, "w", encoding="utf8") as f:
        indent = None if minify else "\t"
        json.dump(gvas_file.dump(), f, indent=indent, cls=CustomEncoder)


def convert_json_to_sav(filename, output_path, force=False):
    print(f"Converting {filename} to SAV, saving to {output_path}")
    if os.path.exists(output_path):
        print(f"{output_path} already exists, this will overwrite the file")
        if not force:
            if not confirm_prompt("Are you sure you want to continue?"):
                exit(1)
    print(f"Loading JSON from {filename}")
    with open(filename, "r", encoding="utf8") as f:
        data = json.load(f)
    gvas_file = GvasFile.load(data)
    print(f"Compressing SAV file")
    if (
        "Pal.PalWorldSaveGame" in gvas_file.header.save_game_class_name
        or "Pal.PalLocalWorldSaveGame" in gvas_file.header.save_game_class_name
    ):
        save_type = 0x32
    else:
        save_type = 0x31
    sav_file = compress_gvas_to_sav(
        gvas_file.write(PALWORLD_CUSTOM_PROPERTIES), save_type
    )
    print(f"Writing SAV file to {output_path}")
    with open(output_path, "wb") as f:
        f.write(sav_file)


def confirm_prompt(question: str) -> bool:
    reply = None
    while reply not in ("y", "n"):
        reply = input(f"{question} (y/n): ").casefold()
    return reply == "y"


if __name__ == "__main__":
    main()



"""BattlEye RCon client."""

from logging import getLogger
from socket import SOCK_DGRAM
from typing import Callable

from rcon.battleye.proto import RESPONSE_TYPES
from rcon.battleye.proto import CommandRequest
from rcon.battleye.proto import Header
from rcon.battleye.proto import LoginRequest
from rcon.battleye.proto import Request
from rcon.battleye.proto import Response
from rcon.battleye.proto import ServerMessage
from rcon.client import BaseClient
from rcon.exceptions import WrongPassword


__all__ = ['Client']


MessageHandler = Callable[[ServerMessage], None]


def log_message(server_message: ServerMessage) -> None:
    """Default handler, logging the server message."""

    getLogger('Server message').info(server_message.message)


class Client(BaseClient, socket_type=SOCK_DGRAM):
    """BattlEye RCon client."""

    def __init__(
            self, *args,
            message_handler: MessageHandler = log_message,
            **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._handle_server_message = message_handler

    def _receive(self, max_length: int) -> Response:
        """Receives a packet."""
        return RESPONSE_TYPES[
            (header := Header.from_bytes(
                (data := self._socket.recv(max_length))[:8]
            )).type
        ].from_bytes(header, data[8:])

    def receive(self, max_length: int = 4096) -> Response:
        """Receives a message."""
        while isinstance(response := self._receive(max_length), ServerMessage):
            self._handle_server_message(response)

        return response

    def communicate(self, request: Request) -> Response:
        """Logs the user in."""
        with self._socket.makefile('wb') as file:
            file.write(bytes(request))

        return self.receive()

    def login(self, passwd: str) -> bool:
        """Logs the user in."""
        if not self.communicate(LoginRequest(passwd)).success:
            raise WrongPassword()

        return True

    def run(self, command: str, *args: str) -> str:
        """Executes a command."""
        return self.communicate(
            CommandRequest.from_command(command, *args)
        ).message



#!/usr/bin/env python3

import psutil
import pprint

def main():
    cpu_cores = []
    phy_cores = []
    phy_core_id = {}
    core_info = {}
    pp = pprint.PrettyPrinter(width=80, compact=True, depth=4)
    with open("/proc/cpuinfo") as fp:
        for line in fp:
            info = line.strip().split(": ")
            if line.strip() == '':
                if int(core_info['core id']) not in phy_cores:
                    phy_cores.append(int(core_info['core id']))
                if int(core_info['core id']) not in phy_core_id:
                    phy_core_id[int(core_info['core id'])] = []
                phy_core_id[int(core_info['core id'])].append(int(core_info['processor']))
                cpu_cores.append(core_info)
                core_info = {}
            else:
                core_info[info[0].strip()] = ''.join(info[1:])
    
    performance_cores = []
    effective_cores = []
    for core_id in phy_core_id:
        if len(phy_core_id[core_id]) > 1:
            performance_cores.append(phy_core_id[core_id][0])
        else:
            effective_cores.append(phy_core_id[core_id][0])
    
    pal_srv = None
    for p in psutil.process_iter():
        if 'PalServer-Linux-Test' in p.name():
            if len(performance_cores) > 0:
                p.cpu_affinity(performance_cores)
            print("PID: %d  %s  RSS: %.2f MB  VMS: %.2f MB  CPU: %s" % (p.pid, p.name(), p.memory_info().rss / 1048576, p.memory_info().vms / 1048576,
                                                                        p.cpu_affinity()))
    
            thread_lists = {}
            for th in p.threads():
                p_th = psutil.Process(th.id)
                thread_lists[p_th.name()] = p_th
                print("  Thread %d: %s  CPU: %s" % (th.id, p_th.name(), p.cpu_affinity()))
            pal_srv = p
            n = 0
            for th_k in sorted(thread_lists.keys()):
                if 'gro-rker' in th_k:
                    thread_lists[th_k].cpu_affinity(performance_cores[n::4])
                    print("  Set Thread %s -> %s" % (th_k, performance_cores[n::4]))
                    n+=1

if __name__ == "__main__":
    main()



from typing import Any, Sequence

from lib.archive import *


def decode(
    reader: FArchiveReader, type_name: str, size: int, path: str
) -> dict[str, Any]:
    if type_name != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {type_name}")
    value = reader.property(type_name, size, path, nested_caller_path=path)
    data_bytes = value["value"]["values"]
    value["value"] = decode_bytes(data_bytes)
    return value


def decode_bytes(c_bytes: Sequence[int]) -> dict[str, Any]:
    if len(c_bytes) == 0:
        return None
    buf = bytes(c_bytes)
    reader = FArchiveReader(buf, debug=False)
    data = {}
    data["id"] = {
        "created_world_id": reader.guid(),
        "local_id_in_created_world": reader.guid(),
        "static_id": reader.fstring(),
    }
    data["type"] = "unknown"
    egg_data = try_read_egg(reader)
    if egg_data != None:
        data |= egg_data
    elif (reader.size - reader.data.tell()) == 4:
        data["type"] = "armor"
        data["durability"] = reader.float()
        if not reader.eof():
            raise Exception("Warning: EOF not reached")
    else:
        cur_pos = reader.data.tell()
        temp_data = {"type": "weapon"}
        try:
            temp_data["durability"] = reader.float()
            temp_data["remaining_bullets"] = reader.i32()
            temp_data["passive_skill_list"] = reader.tarray(lambda r: r.fstring())
            if not reader.eof():
                raise Exception("Warning: EOF not reached")
            data |= temp_data
        except Exception as e:
            print(
                f"Warning: Failed to parse weapon data, continuing as raw data {buf}: {e}"
            )
            reader.data.seek(cur_pos)
            data["trailer"] = [int(b) for b in reader.read_to_end()]
    return data


def try_read_egg(reader: FArchiveReader) -> Optional[dict[str, Any]]:
    cur_pos = reader.data.tell()
    try:
        data = {"type": "egg"}
        data["character_id"] = reader.fstring()
        data["object"] = reader.properties_until_end()
        data["unknown_bytes"] = reader.byte_list(4)
        data["unknown_id"] = reader.guid()
        if not reader.eof():
            raise Exception("Warning: EOF not reached")
        return data
    except Exception as e:
        if e.args[0] == "Warning: EOF not reached":
            raise e
        reader.data.seek(cur_pos)
        return None


def encode(
    writer: FArchiveWriter, property_type: str, properties: dict[str, Any]
) -> int:
    if property_type != "ArrayProperty":
        raise Exception(f"Expected ArrayProperty, got {property_type}")
    del properties["custom_type"]
    encoded_bytes = encode_bytes(properties["value"])
    properties["value"] = {"values": [b for b in encoded_bytes]}
    return writer.property_inner(property_type, properties)


def encode_bytes(p: dict[str, Any]) -> bytes:
    if p is None:
        return bytes()
    writer = FArchiveWriter()
    writer.guid(p["id"]["created_world_id"])
    writer.guid(p["id"]["local_id_in_created_world"])
    writer.fstring(p["id"]["static_id"])
    if p["type"] == "unknown":
        writer.write(bytes(p["trailer"]))
    elif p["type"] == "egg":
        writer.fstring(p["character_id"])
        writer.properties(p["object"])
        writer.write(bytes(p["unknown_bytes"]))
        writer.guid(p["unknown_id"])
    elif p["type"] == "armor":
        writer.float(p["durability"])
    elif p["type"] == "weapon":
        writer.float(p["durability"])
        writer.i32(p["remaining_bullets"])
        writer.tarray(lambda w, d: w.fstring(d), p["passive_skill_list"])
    encoded_bytes = writer.bytes()
    return encoded_bytes
